from __future__ import annotations

import logging
import warnings
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, TypeVar

import numba as nb
import numpy as np
import pandas as pd
import xarray as xr

from larch.warning import WeightRescaleWarning

from .construct import _DatasetConstruct
from .dim_names import ALTID, ALTIDX, CASEALT, CASEID, CASEPTR, GROUPID, INGROUP

if TYPE_CHECKING:
    import sharrow as sh
    from numpy.typing import ArrayLike

    from larch.dataset import DataTree

DataT = TypeVar("DataT", bound=xr.Dataset | xr.DataArray)


@nb.njit
def case_ptr_to_indexes(n_casealts, case_ptrs):
    case_index = np.zeros(n_casealts, dtype=np.int64)
    for c in range(case_ptrs.shape[0] - 1):
        case_index[case_ptrs[c] : case_ptrs[c + 1]] = c
    return case_index


@nb.njit
def any_row_sums_to_not_one(arr):
    for i in range(arr.shape[0]):
        row_sum = 0
        for j in range(arr.shape[1]):
            row_sum += arr[i, j]
        if row_sum < 0.99999999 or row_sum > 1.00000001:
            return True
    return False


@nb.njit
def ce_dissolve_zero_variance(ce_data, ce_caseptr):
    """
    Dissolves zero variance in the given data based on case pointers.

    Parameters
    ----------
    ce_data : array-like, shape [n_casealts]
        One-dimensional array containing the data.
    ce_caseptr : array-like, shape [n_cases]
        One-dimensional array containing the case pointers.

    Returns
    -------
    out : ndarray
        One-dimensional array containing the dissolved data.
    flag : int
        Flag indicating whether variance was detected or not.
        - 1 if variance was detected
        - 0 if no variance was found and the `out` array is valid.
    """
    failed = 0
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:, -1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0],)
    out = np.zeros(shape, dtype=ce_data.dtype)
    c = 0
    out[0] = ce_data[0]
    for row in range(ce_data.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
            out[c] = ce_data[row]
        else:
            if out[c] != ce_data[row]:
                failed = 1
                break
    return out, failed


class _GenericFlow:
    def __init__(self, x: xr.DataArray | xr.Dataset | None = None):
        self._obj = x
        self._flow_library = {}

    @property
    def CASEID(self) -> str | None:
        """Str : The _caseid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEID, None)
        return result

    def __set_attr(self, attr_name, target_value, check_dim=True):
        if target_value is None:
            if attr_name in self._obj.attrs:
                del self._obj.attrs[attr_name]
        else:
            if check_dim and target_value not in self._obj.dims:
                raise ValueError(f"cannot set {attr_name}, {target_value} not in dims")
            self._obj.attrs[attr_name] = target_value

    @CASEID.setter
    def CASEID(self, dim_name):
        self.__set_attr(CASEID, dim_name, check_dim=self.GROUPID is not None)

    @property
    def ALTID(self) -> str | None:
        """The _altid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(ALTID, None)
        return result

    @ALTID.setter
    def ALTID(self, dim_name):
        self.__set_attr(ALTID, dim_name)

    @property
    def CASEALT(self) -> str | None:
        """The _casealt_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEALT, None)
        return result

    @CASEALT.setter
    def CASEALT(self, dim_name):
        self.__set_attr(CASEALT, dim_name)

    @property
    def ALTIDX(self) -> str | None:
        """Str : The _alt_idx_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(ALTIDX, None)
        return result

    @ALTIDX.setter
    def ALTIDX(self, dim_name):
        self.__set_attr(ALTIDX, dim_name, False)

    @property
    def CASEPTR(self) -> str | None:
        """The _caseptr_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(CASEPTR, None)
        return result

    @CASEPTR.setter
    def CASEPTR(self, dim_name):
        self.__set_attr(CASEPTR, dim_name, False)

    @property
    def GROUPID(self) -> str | None:
        """The _groupid_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(GROUPID, None)
        return result

    @GROUPID.setter
    def GROUPID(self, dim_name):
        self.__set_attr(GROUPID, dim_name)

    @property
    def INGROUP(self) -> str | None:
        """The _ingroup_ dimension of this Dataset, if defined."""
        result = self._obj.attrs.get(INGROUP, None)
        return result

    @INGROUP.setter
    def INGROUP(self, dim_name):
        self.__set_attr(INGROUP, dim_name)

    def set_altids(
        self, altids, dim_name=None, dtype=None, inplace=False
    ) -> xr.Dataset:
        """
        Set the alternative ids for this Dataset.

        Parameters
        ----------
        altids : array-like of int
            Integer id codes.
        dim_name : str, optional
            Use this dimension name for alternatives.
        dtype : dtype, optional
            Coerce the altids to be this dtype.
        inplace : bool, default False
            When true, apply the transformation in-place on the Dataset,
            otherwise return a modified copy.

        Returns
        -------
        xarray.Dataset
            The modified Dataset.

        Examples
        --------
        >>> import numpy as np
        >>> from larch.dataset import Dataset
        >>> ds = Dataset()
        >>> ds = ds.dc.set_altids([1,2,3,4], dtype=np.int64)
        >>> ds.dc.altids()
        Int64Index([1, 2, 3, 4], dtype='int64', name='_altid_')
        >>> ds.dc.n_alts
        4
        >>> ds2 = ds.dc.set_altids([7,8,9], dim_name='A', dtype=np.int64)
        >>> ds2.dc.ALTID
        'A'
        >>> ds2.dc.altids()
        Int64Index([7, 8, 9], dtype='int64', name='A')
        >>> ds2
        <xarray.Dataset>
        Dimensions:  (_altid_: 4, A: 3)
        Coordinates:
          * _altid_  (_altid_) int64 1 2 3 4
          * A        (A) int64 7 8 9
        Data variables:
            *empty*
        Attributes:
            _altid_:  A
        """
        if inplace:
            obj = self._obj
        else:
            obj = self._obj.copy()
        dim_name = (
            dim_name or getattr(altids, "name", None) or obj.attrs.get(ALTID, ALTID)
        )
        if not isinstance(altids, xr.DataArray):
            altids = xr.DataArray(
                np.asarray(altids),
                dims=(dim_name),
            )
        if dtype is not None:
            altids = altids.astype(dtype)
        obj.coords[dim_name] = altids
        obj.dc.ALTID = dim_name
        return obj

    def set_altnames(self, altnames, inplace=False) -> xr.Dataset:
        """
        Set the alternative names for this Dataset.

        Parameters
        ----------
        altnames : Mapping or array-like
            A mapping of (integer) codes to names, or an array or names
            of the same length and order as the alternatives already
            defined in this Dataset.
        inplace : bool, default False
            When true, apply the transformation in-place on the Dataset,
            otherwise return a modified copy.

        Returns
        -------
        xarray.Dataset
            The modified Dataset.  A value is returned regardless of whether
            `inplace` is True.
        """
        if inplace:
            obj = self._obj
        else:
            obj = self._obj.copy()
        if isinstance(altnames, Mapping):
            a = obj.dc.ALTID
            names = xr.DataArray(
                [altnames.get(i, None) for i in obj[a].values],
                dims=a,
            )
        elif isinstance(altnames, xr.DataArray):
            names = altnames
        else:
            if obj.dc.ALTID is None:
                obj = obj.dc.set_altids(np.arange(1, len(altnames) + 1))
            names = xr.DataArray(
                np.asarray(altnames),
                dims=obj.dc.ALTID,
            )
        obj.coords["alt_names"] = names
        return obj

    def as_tree(self, label="main", exclude_dims=()) -> DataTree:
        """
        Convert this Dataset to a DataTree.

        For |idco| and |idca| datasets, the result will generally be a
        single-node tree.  For |idce| data, there will be

        Parameters
        ----------
        label : str, default 'main'
            Name to use for the root node in the tree.
        exclude_dims : Tuple[str], optional
            Exclude these dimensions, in addition to any
            dimensions listed in the `_exclude_dims_` attribute.

        Returns
        -------
        DataTree
        """
        from ..dataset import DataTree

        if self.CASEPTR is not None:
            case_index = case_ptr_to_indexes(
                self._obj.sizes[self.CASEALT], self[self.CASEPTR].values
            )
            obj = self._obj.assign(
                {"_case_index_": xr.DataArray(case_index, dims=(self.CASEALT))}
            )
            tree = DataTree(**{label: obj.drop_dims(self.CASEID)})
            ds = obj.keep_dims(self.CASEID)
            ds.attrs.pop("_exclude_dims_", None)
            ds.attrs.pop("_caseptr_", None)
            ds.attrs.pop("_casealt_", None)
            ds.attrs.pop("_alt_idx_", None)
            tree.add_dataset(
                "idcoVars",
                ds,
                relationships=(f"{label}._case_index_ -> idcoVars.{self.CASEID}"),
            )
        else:
            tree = DataTree(**{label: self._obj})
        return tree

    def setup_flow(self, *args, **kwargs) -> sh.Flow:
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        This method creates a new `DataTree` with only this Dataset as
        the root Dataset labeled `main`.  All other arguments are passed
        through to `DataTree.setup_flow`.

        Returns
        -------
        Flow
        """
        return self.as_tree().setup_flow(*args, **kwargs)

    def caseids(self) -> pd.Index:
        """
        Access the caseids coordinates as an index.

        Returns
        -------
        pandas.Index
        """
        return self._obj.indexes[self.CASEID]

    def altids(self) -> pd.Index:
        """
        Access the altids coordinates as an index.

        Returns
        -------
        pandas.Index
        """
        return self._obj.indexes[self.ALTID]

    def groupids(self) -> pd.Index:
        """
        Access the groupids coordinates as an index.

        Returns
        -------
        pandas.Index
        """
        return self._obj.indexes[self.GROUPID]

    @property
    def alts_mapping(self) -> dict[int, str]:
        """Mapping of alternative codes to names."""
        a = self._obj.coords[self.ALTID]
        if "alt_names" in a.coords:
            return dict(zip(a.values, a.coords["alt_names"].values))
        else:
            return dict(zip(a.values, a.values))

    @property
    def n_cases(self) -> int:
        """The number of discrete choice cases in this Dataset."""
        try:
            return self._obj.sizes[self.CASEID]
        except KeyError:
            try:
                return self._obj.sizes[self.GROUPID] * self._obj.sizes[self.INGROUP]
            except KeyError:
                pass
            logging.getLogger().error(
                f"missing {self.CASEID!r} among dims {self._obj.sizes}"
            )
            raise

    @property
    def n_panels(self):
        try:
            return self._obj.sizes[self.GROUPID]
        except KeyError:
            try:
                return self._obj.sizes[self.CASEID]
            except KeyError:
                pass
            logging.getLogger().error(
                f"missing {self.GROUPID!r} and {self.CASEID!r} among dims {self._obj.sizes}"
            )
            raise

    @property
    def n_in_panel(self):
        return self._obj.sizes[self.INGROUP]

    def transfer_dimension_attrs(self, target: DataT) -> DataT:
        """
        Transfer discrete choice dimension attributes to a new Dataset or DataArray.

        Parameters
        ----------
        target : xarray.DataArray or xarray.Dataset

        Returns
        -------
        xarray.DataArray or xarray.Dataset
        """
        if not isinstance(target, xr.DataArray | xr.Dataset):
            return target
        updates = {}
        for i in [CASEID, ALTID, GROUPID, INGROUP, CASEALT, CASEPTR]:
            j = self._obj.attrs.get(i, None)
            if j is not None and j in target.dims:
                updates[i] = j
        return target.assign_attrs(updates)

    def get_expr(self, expression) -> xr.DataArray:
        """
        Access or evaluate an expression.

        Parameters
        ----------
        expression : str

        Returns
        -------
        xarray.DataArray
        """
        try:
            result = self._obj[expression]
        except (KeyError, IndexError):
            if expression in self._flow_library:
                flow = self._flow_library[expression]
            else:
                flow = self.setup_flow({expression: expression})
                self._flow_library[expression] = flow
            if flow.tree.root_dataset is not self:
                flow.tree = self.as_tree()
            result = flow.load_dataarray().isel(expressions=0)
        result = self.transfer_dimension_attrs(result)
        return result

    def dissolve_zero_variance(self, dim="<ALTID>", inplace=False) -> xr.Dataset:
        """
        Dissolve dimension on variables where it has no variance.

        This method is convenient to convert variables that have
        been loaded as |idca| or |idce| format into |idco| format where
        appropriate.

        Parameters
        ----------
        dim : str, optional
            The name of the dimension to potentially dissolve.
        inplace : bool, default False
            Whether to dissolve variables in-place.

        Returns
        -------
        xarray.Dataset
            The modified Dataset. A value is returned regardless of whether
            `inplace` is True.
        """
        if dim == "<ALTID>":
            dim = self.ALTID
        if inplace:
            obj = self
        else:
            obj = self.copy()
        for k in obj.variables:
            if obj[k].dtype.kind in {"U", "S", "O"}:
                continue
            if dim in obj[k].dims:
                try:
                    dissolve = obj[k].std(dim=dim).max() < 1e-10
                except TypeError:
                    pass
                else:
                    if dissolve:
                        obj[k] = obj[k].min(dim=dim)
            elif obj[k].dims == (self.CASEALT,):
                proposal, flag = ce_dissolve_zero_variance(
                    obj[k].values, obj[self.CASEPTR].values
                )
                if flag == 0:
                    obj = obj.assign({k: xr.DataArray(proposal, dims=(self.CASEID))})
        return obj

    def dissolve_coords(self, dim, others=None):
        d = self._obj.reset_index(dim)
        a = d[f"{dim}_"]
        mapper = dict((j, i) for (i, j) in enumerate(a.to_series()))
        mapper_f = np.vectorize(mapper.get)
        if others is None:
            others = []
        if isinstance(others, str):
            others = [others]
        for other in others:
            d[other] = xr.apply_ufunc(mapper_f, d[other])
        return d


@xr.register_dataarray_accessor("dc")
class _DataArrayDC(_GenericFlow):
    """Larch discrete choice attributes and methods for xarray.DataArray."""

    _parent_class = xr.DataArray

    @property
    def n_alts(self):
        """The number of discrete choice alternatives in this DataArray."""
        if self.ALTID in self._obj.sizes:
            return self._obj.sizes.get(self.ALTID)
        if "n_alts" in self._obj.attrs:
            return self._obj.attrs["n_alts"]
        if self.name:
            raise ValueError(f"no n_alts set for {self.name!r}")
        raise ValueError("no n_alts set")

    def __getitem__(self, name):
        # pass dimension attrs to DataArray
        result = self._obj[name]
        result = self.transfer_dimension_attrs(result)
        return result

    def __getattr__(self, name):
        # pass dimension attrs to DataArray
        result = getattr(self._obj, name)
        result = self.transfer_dimension_attrs(result)
        return result


@xr.register_dataset_accessor("dc")
class _DatasetDC(_GenericFlow):
    """Larch discrete choice attributes and methods for xarray.Dataset."""

    _parent_class = xr.Dataset

    @property
    def n_alts(self) -> int:
        """The number of discrete choice alternatives in this Dataset."""
        if self.ALTID in self._obj.sizes:
            return self._obj.sizes[self.ALTID]
        if "n_alts" in self._obj.attrs:
            return self._obj.attrs["n_alts"]
        raise ValueError("no n_alts set")

    def __getitem__(self, name):
        # pass dimension attrs to DataArray
        result = self._obj[name]
        result = self.transfer_dimension_attrs(result)
        return result

    def __getattr__(self, name):
        # pass dimension attrs to DataArray
        result = getattr(self._obj, name)
        result = self.transfer_dimension_attrs(result)
        return result

    def __contains__(self, item):
        return self._obj.__contains__(item)

    def query_cases(self, query, parser="pandas", engine=None):
        """
        Return a new dataset with each array indexed along the CASEID dimension.

        The indexers are given as strings containing Python expressions to be
        evaluated against the data variables in the dataset.

        Parameters
        ----------
        query : str
            Python expressions to be evaluated against the data variables
            in the dataset. The expressions will be evaluated using the pandas
            eval() function, and can contain any valid Python expressions but cannot
            contain any Python statements.
        parser : {"pandas", "python"}, default: "pandas"
            The parser to use to construct the syntax tree from the expression.
            The default of 'pandas' parses code slightly different than standard
            Python. Alternatively, you can parse an expression using the 'python'
            parser to retain strict Python semantics.
        engine : {"python", "numexpr", None}, default: None
            The engine used to evaluate the expression. Supported engines are:

            - None: tries to use numexpr, falls back to python
            - "numexpr": evaluates expressions using numexpr
            - "python": performs operations as if you had eval’d in top level python

        Returns
        -------
        obj : Dataset
            A new Dataset with the same contents as this dataset, except each
            array is indexed by the results of the query on the CASEID dimension.

        See Also
        --------
        Dataset.isel
        pandas.eval
        """
        result = self._obj.query({self.CASEID: query}, parser=parser, engine=engine)
        result = self.transfer_dimension_attrs(result)
        return result

    def to_arrays(self, graph, float_dtype=np.float64):
        from ..model.cascading import array_av_cascade, array_ch_cascade
        from ..model.data_arrays import DataArrays

        if "co" in self:
            co = self["co"].values.astype(float_dtype)
        else:
            co = np.empty((self.n_cases, 0), dtype=float_dtype)

        if "ca" in self:
            ca = self["ca"].values.astype(float_dtype)
        else:
            ca = np.empty((self.n_cases, self.n_alts, 0), dtype=float_dtype)

        if "ce_data" in self:
            ce_data = self["ce_data"].values.astype(float_dtype)
        else:
            ce_data = np.empty((0, 0), dtype=float_dtype)

        if self.ALTIDX is not None:
            ce_altidx = self[self.ALTIDX].values
        else:
            ce_altidx = np.empty((0), dtype=np.int16)

        if self.CASEPTR is not None:
            ce_caseptr = np.lib.stride_tricks.sliding_window_view(
                self[self.CASEPTR].values, 2
            )
        else:
            ce_caseptr = np.empty((self.n_cases, 0), dtype=np.int16)

        if "wt" in self:
            wt = self["wt"].values.astype(float_dtype)
        else:
            if self.GROUPID is None:
                wt = np.ones(self.n_cases, dtype=float_dtype)
            else:
                wt = np.ones(self.n_panels, dtype=float_dtype)

        if "ch" in self:
            ch = array_ch_cascade(self["ch"].values, graph, dtype=float_dtype)
        else:
            if self.GROUPID is None:
                ch = np.zeros([self.n_cases, len(graph)], dtype=float_dtype)
            else:
                ch = np.zeros(
                    [self.n_panels, self.n_in_panel, len(graph)], dtype=float_dtype
                )

        if "av" in self:
            av = array_av_cascade(self["av"].values, graph)
        else:
            if self.GROUPID is None:
                av = np.ones([self.n_cases, len(graph)], dtype=np.int8)
            else:
                av = np.ones(
                    [self.n_panels, self.n_in_panel, len(graph)], dtype=np.int8
                )

        return DataArrays(ch, av, wt, co, ca, ce_data, ce_altidx, ce_caseptr)

    def autoscale_weights(self) -> float:
        """
        Scale the weights so the average weight is 1.

        If weights are embedded in the choice variable,
        they are extracted (so the total choice in each case
        is 0.0 or 1.0, and the weight is isolated in the
        data_wt terms) before any scaling is applied.

        Returns
        -------
        scale : float
        """
        need_to_extract_wgt_from_ch = any_row_sums_to_not_one(self._obj["ch"].values)

        if need_to_extract_wgt_from_ch and "wt" not in self._obj:
            self._obj["wt"] = xr.DataArray(
                data=1.0,
                coords={self.CASEID: self._obj[self.CASEID]},
                dims=(self.CASEID),
            )

        if need_to_extract_wgt_from_ch:
            wgt_from_ch = self._obj["ch"].sum(self.ALTID)
            wgt_from_ch = wgt_from_ch.where(wgt_from_ch > 0, 1)
            self._obj["ch"] /= wgt_from_ch
            self._obj["wt"] *= wgt_from_ch

        if "wt" not in self._obj:
            return 1.0

        total_weight = self._obj["wt"].sum()
        scale_level = float(total_weight) / self.n_cases

        normalization = self._obj["wt"].attrs.get("normalization", 1.0)
        normalization *= scale_level
        self._obj["wt"].attrs["normalization"] = normalization

        if scale_level != 0:
            self._obj["wt"] /= scale_level

        if scale_level < 0.99999 or scale_level > 1.00001:
            warnings.warn(
                f"rescaled array of weights by a factor of {scale_level}",
                WeightRescaleWarning,
                stacklevel=2,
            )

        return scale_level

    def unscale_weights(self) -> float:
        """
        Unscale the weights, restoring the original values.

        This method is the inverse of `autoscale_weights`.

        Returns
        -------
        scale : float
        """
        scale_level = self._obj["wt"].attrs.get("normalization", 1.0)
        if scale_level != 0:
            self._obj["wt"] *= scale_level
        self._obj["wt"].attrs["normalization"] = 1.0
        return scale_level

    @staticmethod
    def from_idco(df, alts: Mapping[int, str] | ArrayLike[int] | None = None):
        """
        Construct a Dataset from an idco-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idco-format DataFrame, with
            the caseid's in a single-level index,
        alts : Mapping or array-like, optional
            If given as a mapping, links alternative codes to names.
            An array or list of integers gives codes for the alternatives,
            which are otherwise unnamed.

        Returns
        -------
        Dataset
        """
        return _DatasetConstruct.from_idco(df, alts=alts)

    @staticmethod
    def from_idca(
        df: pd.DataFrame,
        *,
        crack: bool = True,
        altnames: Mapping[int, str] | Sequence[str] = None,
        avail: str = "_avail_",
        fill_missing: dict = None,
    ) -> xr.Dataset:
        """
        Construct a Dataset from an idca-format DataFrame.

        This method loads the data as dense arrays.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default True
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        avail : str, default '_avail_'
            When the imported data is in idce format (i.e. sparse) then
            an availability indicator is computed and given this name. This
            argument has no effect if the data is already in idca format.
        fill_missing : scalar or Mapping, optional
            Fill values to use for missing values when imported data is
            in idce format (i.e. sparse).  Give a single value to use
            globally, or a mapping of {variable: value} or {dtype: value}.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.dc.from_idce : Construct a Dataset from a sparse idca-format DataFrame.
        """
        return _DatasetConstruct.from_idca(
            df, crack=crack, altnames=altnames, avail=avail, fill_missing=fill_missing
        )

    @staticmethod
    def from_idce(
        df: pd.DataFrame,
        crack: bool = True,
        altnames: Mapping[int, str] | Sequence[str] = None,
        dim_name: str | None = None,
        alt_index: str = "alt_idx",
        case_index: str | None = None,
        case_pointer=None,
    ):
        """
        Construct a Dataset from a sparse idca-format DataFrame.

        Parameters
        ----------
        df : DataFrame
            The input data should be an idca-format or idce-format DataFrame,
            with the caseid's and altid's in a two-level pandas MultiIndex.
        crack : bool, default False
            If True, the `dissolve_zero_variance` method is applied before
            repairing dtypes, to ensure that missing value are handled
            properly.
        altnames : Mapping, optional
            If given as a mapping, links alternative codes to names.
            An array or list of strings gives names for the alternatives,
            sorted in the same order as the codes.
        dim_name : str, optional
            Name to apply to the sparse index dimension.
        alt_index : str, default 'alt_idx'
            Add the alt index (position) for each sparse data row as a
            coords array with this name.
        case_index : str, optional
            Add the case index (position) for each sparse data row as a
            coords array with this name. If not given, this array is not
            stored but it can still be reconstructed later from the case
            pointers.
        case_pointer : str, optional
            Use this name for the case_ptr dimension, overriding the
            default.

        Returns
        -------
        Dataset

        See Also
        --------
        Dataset.from_idca : Construct a dense Dataset from a idca-format DataFrame.
        """
        return _DatasetConstruct.from_idce(
            df,
            crack=crack,
            altnames=altnames,
            dim_name=dim_name,
            alt_index=alt_index,
            case_index=case_index,
            case_pointer=case_pointer,
        )

    def set_dtypes(self, dtypes, inplace=False, on_error="warn"):
        """
        Set the dtypes for the variables in this Dataset.

        Parameters
        ----------
        dtypes : Mapping or DataFrame
            Mapping of names to dtypes, or a DataFrame to infer such a
            mapping.
        inplace : bool, default False
            Whether to convert dtypes inplace.
        on_error : {'warn', 'raise', 'ignore'}
            What to do when a type conversion triggers an error.

        Returns
        -------
        Dataset
        """
        if isinstance(dtypes, pd.DataFrame):
            dtypes = dtypes.dtypes
        if inplace:
            obj = self._obj
        else:
            obj = self._obj.copy()
        for k in obj:
            if k not in dtypes:
                continue
            try:
                obj[k] = obj[k].astype(dtypes[k])
            except Exception as err:
                if on_error == "warn":
                    warnings.warn(
                        f"{err!r} on converting {k} to type {dtypes[k]}", stacklevel=2
                    )
                elif on_error == "raise":
                    raise
        return obj


@xr.register_dataset_accessor("icase")
class _DatasetCaseIslice(_GenericFlow):
    def __getitem__(self, item):
        if self.CASEID is not None:
            return self._obj.isel({self.CASEID: item})
        elif self.GROUPID is not None:
            return self._obj.isel({self.GROUPID: item})
        else:
            raise ValueError("neither CASEID nor GROUPID is defined")



from __future__ import annotations

import warnings
from collections.abc import MutableSequence
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from ..util import dictx
from .basemodel import BaseModel
from .constrainedmodel import ConstrainedModel
from .numbamodel import _arr_inflate, _safe_sqrt
from .possible_overspec import (
    PossibleOverspecification,
    compute_possible_overspecification,
)

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from larch.util import dictx
    from larch.util.excel import ExcelWriter


class ModelGroup(ConstrainedModel, MutableSequence):
    def __init__(self, models, title=None):
        self._submodels = list()
        super().__init__(title=title)
        for model in models:
            if isinstance(model, ModelGroup):
                self._submodels.extend(model._submodels)
            else:
                self._submodels.append(model)
            self._parameter_bucket.attach_model(model)

    def __getitem__(self, x):
        return self._submodels[x]

    def __setitem__(self, i, value):
        assert isinstance(value, BaseModel)
        self._submodels[i] = value
        self._parameter_bucket.attach_model(value)

    def __delitem__(self, x):
        del self._submodels[x]

    def __len__(self):
        return len(self._submodels)

    def __contains__(self, x):
        return x in self._submodels

    def insert(self, i, value):
        assert isinstance(value, BaseModel)
        self._submodels.insert(i, value)
        self._parameter_bucket.attach_model(value)

    def total_weight(self) -> float:
        """
        Compute the total weight across all models in the group.

        Returns
        -------
        float
        """
        return sum([model.total_weight() for model in self._submodels])

    @property
    def n_cases(self):
        """Int : Total number of cases in all submodels."""
        return sum([model.n_cases for model in self._submodels])

    def logloss(
        self,
        x: ArrayLike | dict | None = None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
        check_if_best=True,
    ) -> float:
        result = self.loglike(x, check_if_best=check_if_best)
        return -result / self.total_weight()

    def d_logloss(self, x=None, start_case=0, stop_case=-1, step_case=1, **kwargs):
        result = self.d_loglike(x, **kwargs)
        return -np.asarray(result) / self.total_weight()

    def loglike(
        self, x: ArrayLike | dict | None = None, *, check_if_best=True, **kwargs
    ) -> float:
        """
        Compute the log likelihood of the group of models.

        Note that unlike for a single model, the log likelihood cannot
        be computed for a slice of cases, but must be computed for all
        cases.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.
        check_if_best : bool, default True
            If True, check if the current log likelihood is the best
            found so far, and if so, update the cached best log likelihood
            and cached best parameters.

        Returns
        -------
        float
        """
        result = sum(
            [
                model.loglike(x=x, check_if_best=False, **kwargs)
                for model in self._submodels
            ]
        )
        if check_if_best:
            self._check_if_best(result)
        return result

    def d_loglike(
        self,
        x=None,
        *,
        return_series=False,
        **kwargs,
    ):
        """
        Compute the gradient of the log likelihood of the group of models.

        Note that unlike for a single model, the gradient of the log
        likelihood cannot be computed for a slice of cases, but must be
        computed for all cases.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the gradient of the log likelihood.  If given as array-like,
            the array must be a vector with length equal to the length of
            the parameter frame, and the given vector will replace the
            current values.  If given as a dictionary, the dictionary is
            used to update the parameters.

        Returns
        -------
        array-like
        """
        return sum(
            [
                model.d_loglike(x=x, return_series=return_series, **kwargs)
                for model in self._submodels
            ]
        )

    def loglike_casewise(
        self,
        x=None,
        **kwargs,
    ):
        """
        Compute the log likelihood case-by-case for the group of models.

        Parameters
        ----------
        x : array-like or dict, optional
            New values to set for the parameters before evaluating
            the log likelihood.  If given as array-like, the array must
            be a vector with length equal to the length of the
            parameter frame, and the given vector will replace
            the current values.  If given as a dictionary,
            the dictionary is used to update the parameters.

        Returns
        -------
        array-like
        """
        return np.concatenate(
            [model.loglike_casewise(x=x, **kwargs) for model in self._submodels]
        )

    def loglike_null(self, use_cache=True):
        """
        Compute the log likelihood at null values.

        Set all parameter values to the value indicated in the
        "nullvalue" column of the parameter frame, and compute
        the log likelihood with the currently loaded data.  Note
        that the null value for each parameter may not be zero
        (for example, the default null value for logsum parameters
        in a nested logit model is 1).

        Parameters
        ----------
        use_cache : bool, default True
            Use the cached value if available.  Set to -1 to
            raise an exception if there is no cached value.

        Returns
        -------
        float
        """
        return sum([model.loglike_null(use_cache) for model in self._submodels])

    def bhhh(
        self,
        x=None,
        *,
        return_dataframe=False,
        **kwargs,
    ):
        result = sum(
            [
                model.bhhh(x=x, return_dataframe=False, **kwargs)
                for model in self._submodels
            ]
        )
        if return_dataframe:
            result = pd.DataFrame(result, columns=self.pnames, index=self.pnames)
        return result

    def loglike2_bhhh(
        self,
        x=None,
        *,
        return_series=False,
        **kwargs,  # swallow any extra arguments
    ):
        result = dictx()
        for model in self._submodels:
            t = model.loglike2_bhhh(
                x=x,
                return_series=return_series,
                start_case=None,
                stop_case=None,
                step_case=None,
                persist=0,
            )
            for k, v in t.items():
                if k in result:
                    result[k] += v
                else:
                    result[k] = v
        return result

    def fit_bhhh(self, *args, **kwargs):
        from .numba_optimization import fit_bhhh

        return fit_bhhh(self, *args, **kwargs)

    def mangle(self, data=True, structure=True):
        for model in self._submodels:
            model.mangle(data, structure)

    def unmangle(self, force=False, structure_only=False):
        for model in self._submodels:
            model.unmangle(force, structure_only)

    def reflow_data_arrays(self) -> None:
        """Reload the internal data_arrays so they are consistent with the datatree."""
        for model in self._submodels:
            model.reflow_data_arrays()

    def doctor(self, **kwargs):
        diagnosis = []
        check_overspec = kwargs.pop("check_overspec", None)
        if check_overspec and len(kwargs) == 0:
            # when only check_overspec is passed, set verbose to 3 to signal
            # that no other checks are requested
            kwargs = {"verbose": 3}
        for model in self._submodels:
            diagnosis.append(model.doctor(**kwargs))
        from .troubleshooting import overspecification

        if check_overspec:
            self, overspec_diagnosis = overspecification(self, check_overspec)
            if overspec_diagnosis:
                diagnosis.append(overspec_diagnosis)
        return self, diagnosis

    def maximize_loglike(
        self,
        *args,
        **kwargs,
    ):
        """
        Maximize the log likelihood.

        Parameters
        ----------
        method : str, optional
            The optimization method to use.  See scipy.optimize for
            most possibilities, or use 'BHHH'. Defaults to SLSQP if
            there are any constraints or finite parameter bounds,
            otherwise defaults to BHHH.
        quiet : bool, default False
            Whether to suppress the dashboard.
        options : dict, optional
            These options are passed through to the `scipy.optimize.minimize`
            function.
        maxiter : int, optional
            Maximum number of iterations.  This argument is just added to
            `options` for most methods.

        Returns
        -------
        dictx
            A dictionary of results, including final log likelihood,
            elapsed time, and other statistics.  The exact items
            included in output will vary by estimation method.

        """
        from .optimization import maximize_loglike

        return maximize_loglike(self, *args, **kwargs)

    def estimate(self, *args, **kwargs):
        """
        Maximize loglike, and then calculate parameter covariance.

        This convenience method runs the following methods in order:
        - maximize_loglike
        - calculate_parameter_covariance

        All arguments are passed through to maximize_loglike.

        Returns
        -------
        dictx
        """
        result = self.maximize_loglike(*args, **kwargs)
        self.calculate_parameter_covariance()
        return result

    def d2_loglike(
        self,
        x=None,
        *,
        start_case=None,
        stop_case=None,
        step_case=None,
    ):
        """
        Compute the Hessian matrix of the log likelihood.

        Parameters
        ----------
        x : array-like, optional
            The parameter values to use in the calculation.  If not
            given, the current parameter values are used.
        start_case : int, optional
            The first case to include in the calculation.
        stop_case : int, optional
            The last case to include in the calculation.
        step_case : int, optional
            The step size between cases to include in the calculation.

        Returns
        -------
        array
        """
        if x is None:
            x = self.pvals.copy()
        from ..util.math import approx_fprime

        return approx_fprime(
            x,
            lambda y: self.d_loglike(
                y,
                start_case=start_case,
                stop_case=stop_case,
                step_case=step_case,
            ),
        )

    def calculate_parameter_covariance(self, pvals=None, *, robust=False):
        """
        Calculate the parameter covariance matrix.

        Parameters
        ----------
        pvals : array-like, optional
            The parameter values to use in the calculation.  If not
            given, the current parameter values are used.
        robust : bool, default False
            Whether to calculate the robust covariance matrix.

        Returns
        -------
        se : array
            The standard errors of the parameter estimates.
        hess : array
            The Hessian matrix.
        ihess : array
            The inverse of the Hessian matrix.
        """
        if pvals is None:
            pvals = self.pvals
        locks = np.asarray(self.pholdfast.astype(bool))
        if self.compute_engine == "jax":
            se, hess, ihess = self.jax_param_cov(pvals)
        else:
            hess = -self.d2_loglike(pvals)
            if self.parameters["holdfast"].sum():
                free = self.pholdfast == 0
                hess_ = hess[free][:, free]
                try:
                    ihess_ = np.linalg.inv(hess_)
                except np.linalg.LinAlgError:
                    ihess_ = np.linalg.pinv(hess_)
                ihess = _arr_inflate(ihess_, locks)
            else:
                try:
                    ihess = np.linalg.inv(hess)
                except np.linalg.LinAlgError:
                    ihess = np.linalg.pinv(hess)
            se = _safe_sqrt(ihess.diagonal())
            self.pstderr = se
        hess = np.asarray(hess).copy()
        hess[locks, :] = 0
        hess[:, locks] = 0
        ihess = np.asarray(ihess).copy()
        ihess[locks, :] = 0
        ihess[:, locks] = 0
        self.add_parameter_array("hess", hess)
        self.add_parameter_array("ihess", ihess)

        overspec = compute_possible_overspecification(hess, self.pholdfast)
        if overspec:
            warnings.warn(
                "Model is possibly over-specified (hessian is nearly singular).",
                category=PossibleOverspecification,
                stacklevel=2,
            )
            possible_overspecification = []
            for eigval, ox, eigenvec in overspec:
                if eigval == "LinAlgError":
                    possible_overspecification.append((eigval, [ox], [""]))
                else:
                    paramset = list(np.asarray(self.pnames)[ox])
                    possible_overspecification.append((eigval, paramset, eigenvec[ox]))
            self._possible_overspecification = possible_overspecification

        # constrained covariance
        if self.constraints:
            constraints = list(self.constraints)
        else:
            constraints = []
        try:
            constraints.extend(self._get_bounds_constraints())
        except Exception:
            pass

        if constraints:
            binding_constraints = list()
            self.add_parameter_array("unconstrained_std_err", self.pstderr)
            self.add_parameter_array("unconstrained_covariance_matrix", ihess)

            s = np.asarray(ihess)
            pvals = self.pvals
            for c in constraints:
                if np.absolute(c.fun(pvals)) < c.binding_tol:
                    binding_constraints.append(c)
                    b = c.jac(self.pf.value)
                    den = b @ s @ b
                    if den != 0:
                        s = s - (1 / den) * s @ b.reshape(-1, 1) @ b.reshape(1, -1) @ s
            self.add_parameter_array("covariance_matrix", s)
            self.pstderr = _safe_sqrt(s.diagonal())

            # Fix numerical issues on some constraints, add constrained notes
            if binding_constraints or any(self.pholdfast != 0):
                notes = {}
                for c in binding_constraints:
                    pa = c.get_parameters()
                    for p in pa:
                        # if self.pf.loc[p, 't_stat'] > 1e5:
                        #     self.pf.loc[p, 't_stat'] = np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        # if self.pf.loc[p, 't_stat'] < -1e5:
                        #     self.pf.loc[p, 't_stat'] = -np.inf
                        #     self.pf.loc[p, 'std_err'] = np.nan
                        n = notes.get(p, [])
                        n.append(c.get_binding_note(pvals))
                        notes[p] = n
                constrained_note = (
                    pd.Series({k: "\n".join(v) for k, v in notes.items()}, dtype=object)
                    .reindex(self.pnames)
                    .fillna("")
                )
                constrained_note[self.pholdfast != 0] = "fixed value"
                self.add_parameter_array("constrained", constrained_note)

        if robust:
            raise NotImplementedError(
                "Robust covariance not yet implemented for ModelGroup."
            )
            # self.robust_covariance()
            # se = self.parameters["robust_std_err"]

        return se, hess, ihess

    def to_xlsx(
        self,
        filename,
        save_now=True,
        data_statistics: bool = True,
        nesting: bool = True,
        embed_model: bool = True,
    ) -> ExcelWriter:
        """
        Write the estimation results to an Excel file.

        Parameters
        ----------
        filename : str
            The name of the file to write.
        save_now : bool, default True
            Whether to save the file immediately.  If False, the
            ExcelWriter object is returned.
        data_statistics : bool, default True
            Whether to include data statistics in the Excel file.
        nesting : bool, default True
            Whether to include nesting statistics in the Excel file.
        embed_model : bool, default True
            Whether to embed the model in the Excel file.

        Returns
        -------
        larch.util.excel.ExcelWriter or None
        """
        from larch.util.excel import _make_excel_writer

        result = _make_excel_writer(self, filename, save_now=False)
        result._post_init(
            filename,
            model=self,
            data_statistics=data_statistics,
            nesting=nesting,
            embed=embed_model,
        )
        if save_now:
            result.close()
        else:
            return result



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    d = d.sel(caseid=d.femdum == 0)

    m = larch.Model(d, compute_engine="numba")

    from larch import P, X

    for a in [4, 5, 6]:
        m.utility_co[a] += X("hhinc") * P(f"hhinc#{a}")

    for i in d["alt_names"][1:3]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_SR")
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    for i in d["alt_names"][3:]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_" + name)
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    m.utility_ca = (
        +X("totcost/hhinc") * P("costbyincome")
        + X("tottime * (altid <= 4)") * P("motorized_time")
        + X("tottime * (altid >= 5)") * P("nonmotorized_time")
        + X("ovtt/dist * (altid <= 4)") * P("motorized_ovtbydist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 17C, Segmented for males"

    m.choice_avail_summary()

    m.set_cap(25)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, options={"maxiter": 1000, "ftol": 1e-10})
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "Household",
            "hhinc.*",
            "vehbywrk.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -1.928007825476385,
        "ASC_SR2": -1.912455790177904,
        "ASC_SR3+": -3.5514203840398006,
        "ASC_Transit": -0.8658094170164065,
        "ASC_Walk": -1.2203008866494844,
        "costbyincome": -0.06390233682221723,
        "hhinc#4": -0.0020804355665526397,
        "hhinc#5": -0.0013784543619515673,
        "hhinc#6": -0.00501951838450115,
        "motorized_ovtbydist": -0.18639734021256957,
        "motorized_time": -0.01951022691589421,
        "nonmotorized_time": -0.024446076816243063,
        "vehbywrk_Bike": -0.9930978890574966,
        "vehbywrk_SR": -0.20981196496890442,
        "vehbywrk_Transit": -0.8322463141471179,
        "vehbywrk_Walk": -0.6105805888289391,
        "wkcbd_Bike": 0.3204533731833982,
        "wkcbd_SR2": 0.027661592161988033,
        "wkcbd_SR3+": 1.423466120894847,
        "wkcbd_Transit": 1.196921606447681,
        "wkcbd_Walk": 0.22281387280820592,
        "wkempden_Bike": 0.0004958705725495997,
        "wkempden_SR2": 0.0009305263934769419,
        "wkempden_SR3+": 0.0005787415453590467,
        "wkempden_Transit": 0.002487830750207477,
        "wkempden_Walk": 0.0012923574145590863,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 2e-2
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import numpy


class PossibleOverspecification(Warning):
    pass


class PossibleOverspecificationError(ValueError):
    pass


def compute_possible_overspecification(
    a, holdfast_vector=None, eigenvalue_threshold=0.005, participation_magnitude=5
):
    """
    Compute the possible overspecification of a model from its hessian matrix.

    Parameters
    ----------
    a : matrix
        The hessian matrix
    holdfast_vector : vector
        A vector of indicators for which row/cols should be ignored as holdfast-ed
    eigenvalue_threshold : float, default 0.005
        The threshold to consider the absolute value of an eigenvalue as close
        enough to zero to be considered as a possible overspecification.  This
        should be a small positive number.
    participation_magnitude : int, default 5
        The magnitude of the participation to consider a parameter as
        participating in a possible overspecification.  Each problematic eigenvector
        is rounded to this number of decimal places to determine the participation
        of each parameter (any non-zero value is considered participating).

    Returns
    -------
    list of tuples
        A list of possible overspecification problems in the model.  Each problem
        is a tuple containing the eigenvalue, the indices of the non-zero elements
        in the eigenvector, and the eigenvector itself.
    """
    ret = []
    if holdfast_vector is None:
        holdfast_vector = numpy.zeros(a.shape[0], dtype=bool)
    else:
        holdfast_vector = holdfast_vector.astype(bool, copy=True)
    holdfast_vector |= (a == 0).all(0)
    a_packed = a[~holdfast_vector, :][:, ~holdfast_vector]
    try:
        eigenvalues_packed, eigenvectors_packed = numpy.linalg.eigh(a_packed)
    except numpy.linalg.linalg.LinAlgError as err:
        return [("LinAlgError", str(err), "")]
    for i in range(len(eigenvalues_packed)):
        if numpy.abs(eigenvalues_packed[i]) < eigenvalue_threshold:
            v = eigenvectors_packed[:, i]
            v = numpy.round(v, participation_magnitude)
            v_unpacked = numpy.zeros(a.shape[0])
            v_unpacked[~holdfast_vector.astype(bool)] = v
            ret.append((eigenvalues_packed[i], numpy.where(v_unpacked)[0], v_unpacked))
    return ret



from __future__ import annotations

import datetime
import os
import shutil
from collections.abc import Collection
from pathlib import Path

import pandas as pd
import xarray as xr
import yaml

from .dataset import construct, from_zarr_with_attr
from .relationships import DataTree, Relationship


def timestamp():
    return datetime.datetime.now(datetime.timezone.utc).astimezone().isoformat()


class ReadOnlyError(ValueError):
    """Object is read-only."""


def _read_parquet(filename, index_col=None) -> xr.Dataset:
    import pyarrow.parquet as pq

    from sharrow.dataset import from_table

    content = pq.read_table(filename)
    if index_col is not None:
        index = content.column(index_col)
        content = content.drop([index_col])
    else:
        index = None
    x = from_table(content, index=index, index_name=index_col or "index")
    return x


class DataStore:
    metadata_filename: str = "metadata.yaml"
    checkpoint_subdir: str = "checkpoints"
    LATEST = "_"
    _BY_OFFSET = "digitizedOffset"

    def __init__(self, directory: Path | None, mode="a", storage_format: str = "zarr"):
        self._directory = Path(directory) if directory else directory
        self._mode = mode
        self._checkpoints = {}
        self._checkpoint_order = []
        self._tree = DataTree(root_node_name=False)
        self._keep_digitized = False
        assert storage_format in {"zarr", "parquet", "hdf5"}
        self._storage_format = storage_format
        try:
            self.read_metadata()
        except FileNotFoundError:
            pass

    @property
    def directory(self) -> Path:
        if self._directory is None:
            raise NotADirectoryError("no directory set")
        return self._directory

    def __setitem__(self, key: str, value: xr.Dataset | pd.DataFrame):
        assert isinstance(key, str)
        if self._mode == "r":
            raise ReadOnlyError
        if isinstance(value, xr.Dataset):
            self._set_dataset(key, value)
        elif isinstance(value, pd.DataFrame):
            self._set_dataset(key, construct(value))
        else:
            raise TypeError(f"cannot put {type(value)}")

    def __getitem__(self, key: str):
        assert isinstance(key, str)
        return self.get_dataset(key)

    def clone(self, mode="a"):
        """
        Create a clone of this DataStore.

        The clone has the same active datasets as the original (and the data
        shares the same memory) but it does not retain the checkpoint metadata
        and is not connected to the same checkpoint store.

        Returns
        -------
        DataStore
        """
        duplicate = self.__class__(None, mode=mode)
        duplicate._tree = self._tree
        return duplicate

    def _set_dataset(
        self,
        name: str,
        data: xr.Dataset,
        last_checkpoint: str = None,
    ) -> None:
        if self._mode == "r":
            raise ReadOnlyError
        data_vars = {}
        coords = {}
        for k, v in data.coords.items():
            coords[k] = v.assign_attrs(last_checkpoint=last_checkpoint)
        for k, v in data.items():
            if k in coords:
                continue
            data_vars[k] = v.assign_attrs(last_checkpoint=last_checkpoint)
        data = xr.Dataset(data_vars=data_vars, coords=coords, attrs=data.attrs)
        self._tree.add_dataset(name, data)

    def _update_dataset(
        self,
        name: str,
        data: xr.Dataset,
        last_checkpoint=None,
    ) -> xr.Dataset:
        if self._mode == "r":
            raise ReadOnlyError
        if not isinstance(data, xr.Dataset):
            raise TypeError(type(data))
        partial_update = self._tree.get_subspace(name, default_empty=True)
        for k, v in data.items():
            if k in data.coords:
                continue
            assert v.name == k
            partial_update = self._update_dataarray(
                name, v, last_checkpoint, partial_update=partial_update
            )
        for k, v in data.coords.items():
            assert v.name == k
            partial_update = self._update_dataarray(
                name, v, last_checkpoint, as_coord=True, partial_update=partial_update
            )
        return partial_update

    def _update_dataarray(
        self,
        name: str,
        data: xr.DataArray,
        last_checkpoint=None,
        as_coord=False,
        partial_update=None,
    ) -> xr.Dataset:
        if self._mode == "r":
            raise ReadOnlyError
        if partial_update is None:
            base_data = self._tree.get_subspace(name, default_empty=True)
        else:
            base_data = partial_update
        if isinstance(data, xr.DataArray):
            if as_coord:
                updated_dataset = base_data.assign_coords(
                    {data.name: data.assign_attrs(last_checkpoint=last_checkpoint)}
                )
            else:
                updated_dataset = base_data.assign(
                    {data.name: data.assign_attrs(last_checkpoint=last_checkpoint)}
                )
            self._tree = self._tree.replace_datasets(
                {name: updated_dataset}, redigitize=self._keep_digitized
            )
            return updated_dataset
        else:
            raise TypeError(type(data))

    def update(
        self,
        name: str,
        obj: xr.Dataset | xr.DataArray,
        last_checkpoint: str = None,
    ) -> None:
        """
        Make a partial update of an existing named dataset.

        Parameters
        ----------
        name : str
        obj : Dataset or DataArray
        last_checkpoint : str or None
            Set the "last_checkpoint" attribute on all updated variables to this
            value.  Users should typically leave this as "None", which flags the
            checkpointing algorithm to write this data to disk the next time a
            checkpoint is written.
        """
        if isinstance(obj, xr.Dataset):
            self._update_dataset(name, obj, last_checkpoint=last_checkpoint)
        elif isinstance(obj, xr.DataArray):
            self._update_dataarray(name, obj, last_checkpoint=last_checkpoint)
        else:
            raise TypeError(type(obj))

    def set_data(
        self,
        name: str,
        data: xr.Dataset | pd.DataFrame,
        relationships: str | Relationship | Collection[str | Relationship] = None,
    ) -> None:
        """
        Set the content of a named dataset.

        This completely overwrites any existing data with the same name.

        Parameters
        ----------
        name : str
        data : Dataset or DataFrame
        relationships : str or Relationship or list thereof
        """
        self.__setitem__(name, data)
        if relationships is not None:
            if isinstance(relationships, (str, Relationship)):
                relationships = [relationships]
            for r in relationships:
                self._tree.add_relationship(r)

    def get_dataset(self, name: str, columns: Collection[str] = None) -> xr.Dataset:
        """
        Retrieve some or all of a named dataset.

        Parameters
        ----------
        name : str
        columns : Collection[str], optional
            Get only these variables of the dataset.
        """
        if columns is None:
            return self._tree.get_subspace(name)
        else:
            return xr.Dataset({c: self._tree[f"{name}.{c}"] for c in columns})

    def get_dataframe(self, name: str, columns: Collection[str] = None) -> pd.DataFrame:
        """
        Retrieve some or all of a named dataset, as a pandas DataFrame.

        This completely overwrites any existing data with the same name.

        Parameters
        ----------
        name : str
        columns : Collection[str], optional
            Get only these variables of the dataset.
        """
        dataset = self.get_dataset(name, columns)
        return dataset.single_dim.to_pandas()

    def _to_be_checkpointed(self) -> dict[str, xr.Dataset]:
        result = {}
        for table_name, table_data in self._tree.subspaces_iter():
            # any data elements that were created without a
            # last_checkpoint attr get one now
            for _k, v in table_data.variables.items():
                if "last_checkpoint" not in v.attrs:
                    v.attrs["last_checkpoint"] = None
            # collect everything not checkpointed
            uncheckpointed = table_data.filter_by_attrs(last_checkpoint=None)
            if uncheckpointed:
                result[table_name] = uncheckpointed
        return result

    def _zarr_subdir(self, table_name, checkpoint_name):
        return self.directory.joinpath(table_name, checkpoint_name).with_suffix(".zarr")

    def _parquet_name(self, table_name, checkpoint_name):
        return self.directory.joinpath(table_name, checkpoint_name).with_suffix(
            ".parquet"
        )

    def make_checkpoint(self, checkpoint_name: str, overwrite: bool = True):
        """
        Write data to disk.

        Only new data (since the last time a checkpoint was made) is actually
        written out.

        Parameters
        ----------
        checkpoint_name : str
        overwrite : bool, default True
        """
        if self._mode == "r":
            raise ReadOnlyError
        to_be_checkpointed = self._to_be_checkpointed()
        new_checkpoint = {
            "timestamp": timestamp(),
            "tables": {},
            "relationships": [],
        }
        # remove checkpoint name from ordered list if it already exists
        while checkpoint_name in self._checkpoint_order:
            self._checkpoint_order.remove(checkpoint_name)
        # add checkpoint name at end ordered list
        self._checkpoint_order.append(checkpoint_name)
        for table_name, table_data in to_be_checkpointed.items():
            if self._storage_format == "parquet" and len(table_data.dims) == 1:
                target = self._parquet_name(table_name, checkpoint_name)
                if overwrite and target.is_file():
                    os.unlink(target)
                target.parent.mkdir(parents=True, exist_ok=True)
                table_data.single_dim.to_parquet(str(target))
            elif self._storage_format == "zarr" or (
                self._storage_format == "parquet" and len(table_data.dims) > 1
            ):
                # zarr is used if ndim > 1
                target = self._zarr_subdir(table_name, checkpoint_name)
                if overwrite and target.is_dir():
                    shutil.rmtree(target)
                target.mkdir(parents=True, exist_ok=True)
                table_data.to_zarr_with_attr(target)
            elif self._storage_format == "hdf5":
                raise NotImplementedError
            else:
                raise ValueError(
                    f"cannot write with storage format {self._storage_format!r}"
                )
            self.update(table_name, table_data, last_checkpoint=checkpoint_name)
        for table_name, table_data in self._tree.subspaces_iter():
            inventory = {"data_vars": {}, "coords": {}}
            for varname, vardata in table_data.items():
                inventory["data_vars"][varname] = {
                    "last_checkpoint": vardata.attrs.get("last_checkpoint", "MISSING"),
                    "dtype": str(vardata.dtype),
                }
            for varname, vardata in table_data.coords.items():
                _cp = checkpoint_name
                # coords in every checkpoint with any content
                if table_name not in to_be_checkpointed:
                    _cp = vardata.attrs.get("last_checkpoint", "MISSING")
                inventory["coords"][varname] = {
                    "last_checkpoint": _cp,
                    "dtype": str(vardata.dtype),
                }
            new_checkpoint["tables"][table_name] = inventory
        for r in self._tree.list_relationships():
            new_checkpoint["relationships"].append(r.to_dict())
        self._checkpoints[checkpoint_name] = new_checkpoint
        self._write_checkpoint(checkpoint_name, new_checkpoint)
        self._write_metadata()

    def _write_checkpoint(self, name, checkpoint):
        if self._mode == "r":
            raise ReadOnlyError
        checkpoint_metadata_target = self.directory.joinpath(
            self.checkpoint_subdir, f"{name}.yaml"
        )
        if checkpoint_metadata_target.exists():
            n = 1
            while checkpoint_metadata_target.with_suffix(f".{n}.yaml").exists():
                n += 1
            os.rename(
                checkpoint_metadata_target,
                checkpoint_metadata_target.with_suffix(f".{n}.yaml"),
            )
        checkpoint_metadata_target.parent.mkdir(parents=True, exist_ok=True)
        with open(checkpoint_metadata_target, "w") as f:
            yaml.safe_dump(checkpoint, f)

    def _write_metadata(self):
        if self._mode == "r":
            raise ReadOnlyError
        metadata_target = self.directory.joinpath(self.metadata_filename)
        if metadata_target.exists():
            n = 1
            while metadata_target.with_suffix(f".{n}.yaml").exists():
                n += 1
            os.rename(metadata_target, metadata_target.with_suffix(f".{n}.yaml"))
        with open(metadata_target, "w") as f:
            metadata = dict(
                datastore_format_version=1,
                checkpoint_order=self._checkpoint_order,
            )
            yaml.safe_dump(metadata, f)

    def read_metadata(self, checkpoints=None):
        """
        Read storage metadata.

        Parameters
        ----------
        checkpoints : str | list[str], optional
            Read only these checkpoints.  If not provided, only the latest
            checkpoint metadata is read. Set to "*" to read all.
        """
        with open(self.directory.joinpath(self.metadata_filename)) as f:
            metadata = yaml.safe_load(f)
        datastore_format_version = metadata.get("datastore_format_version", "missing")
        if datastore_format_version == 1:
            self._checkpoint_order = metadata["checkpoint_order"]
        else:
            raise NotImplementedError(f"{datastore_format_version=}")
        if checkpoints is None or checkpoints == self.LATEST:
            checkpoints = [self._checkpoint_order[-1]]
        elif isinstance(checkpoints, str):
            if checkpoints == "*":
                checkpoints = list(self._checkpoint_order)
            else:
                checkpoints = [checkpoints]
        for c in checkpoints:
            with open(
                self.directory.joinpath(self.checkpoint_subdir, f"{c}.yaml")
            ) as f:
                self._checkpoints[c] = yaml.safe_load(f)

    def restore_checkpoint(self, checkpoint_name: str):
        if checkpoint_name not in self._checkpoints:
            try:
                self.read_metadata(checkpoint_name)
            except FileNotFoundError:
                raise KeyError(checkpoint_name) from None
        checkpoint = self._checkpoints[checkpoint_name]
        self._tree = DataTree(root_node_name=False)
        for table_name, table_def in checkpoint["tables"].items():
            if table_name == "timestamp":
                continue
            t = xr.Dataset()
            opened_targets = {}
            coords = table_def.get("coords", {})
            if len(coords) == 1:
                index_name = list(coords)[0]
            else:
                index_name = None
            for coord_name, coord_def in coords.items():
                target = self._zarr_subdir(table_name, coord_def["last_checkpoint"])
                if target.exists():
                    if target not in opened_targets:
                        opened_targets[target] = from_zarr_with_attr(target)
                else:
                    # zarr not found, try parquet
                    target2 = self._parquet_name(
                        table_name, coord_def["last_checkpoint"]
                    )
                    if target2.exists():
                        if target not in opened_targets:
                            opened_targets[target] = _read_parquet(target2, index_name)
                    else:
                        raise FileNotFoundError(target)
                t = t.assign_coords({coord_name: opened_targets[target][coord_name]})
            data_vars = table_def.get("data_vars", {})
            for var_name, var_def in data_vars.items():
                if var_def["last_checkpoint"] == "MISSING":
                    raise ValueError(f"missing checkpoint for {table_name}.{var_name}")
                target = self._zarr_subdir(table_name, var_def["last_checkpoint"])
                if target.exists():
                    if target not in opened_targets:
                        opened_targets[target] = from_zarr_with_attr(target)
                else:
                    # zarr not found, try parquet
                    target2 = self._parquet_name(table_name, var_def["last_checkpoint"])
                    if target2.exists():
                        if target not in opened_targets:
                            opened_targets[target] = _read_parquet(target2, index_name)
                    else:
                        raise FileNotFoundError(target)
                t = t.assign({var_name: opened_targets[target][var_name]})
            self._tree.add_dataset(table_name, t)
        for r in checkpoint["relationships"]:
            self._tree.add_relationship(Relationship(**r))

    def add_relationship(self, relationship: str | Relationship):
        self._tree.add_relationship(relationship)

    def digitize_relationships(self, redigitize=True):
        """
        Convert all label-based relationships into position-based.

        Parameters
        ----------
        redigitize : bool, default True
            Re-compute position-based relationships from labels, even
            if the relationship had previously been digitized.
        """
        self._keep_digitized = True
        self._tree.digitize_relationships(inplace=True, redigitize=redigitize)

    @property
    def relationships_are_digitized(self) -> bool:
        """Bool : Whether all relationships are digital (by position)."""
        return self._tree.relationships_are_digitized



import base64
import contextlib
import filecmp
import os
import uuid

from filelock import FileLock, Timeout


def blacken(code):
    try:
        import black
    except ImportError:
        return code
    mode = black.FileMode()
    fast = False
    try:
        return black.format_file_contents(code, fast=fast, mode=mode)
    except black.NothingChanged:
        return code
    except Exception as err:
        import warnings

        warnings.warn(f"error in blacken: {err!r}", stacklevel=2)
        return code


@contextlib.contextmanager
def rewrite(pth, mode="wt"):
    if not os.path.exists(pth):
        with open(pth, mode) as f:
            yield f
    else:
        unique_id = base64.b32encode(uuid.uuid4().bytes)[:26].decode()
        pth_, pth_ext = os.path.splitext(pth)
        temp_file = pth_ + "_temp_" + unique_id + pth_ext
        with open(temp_file, mode) as f:
            yield f
        lock_path = pth + ".lock"
        lock = FileLock(lock_path, timeout=15)
        try:
            with lock.acquire():
                if filecmp.cmp(pth, temp_file, shallow=False):
                    # no changes, delete temp_file
                    os.remove(temp_file)
                else:
                    # changes, delete original and move tempfile
                    os.remove(pth)
                    os.rename(temp_file, pth)
        except Timeout:
            pass



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    d = d.sel(caseid=d.numveh >= 2)

    m = larch.Model(d, compute_engine="numba")

    from larch import P, X

    for a in [4, 5, 6]:
        m.utility_co[a] += X("hhinc") * P(f"hhinc#{a}")

    for i in d["alt_names"][1:3]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_SR")
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    for i in d["alt_names"][3:]:
        name = str(i.values)
        a = int(i.altid)
        m.utility_co[a] += (
            +X("vehbywrk") * P("vehbywrk_" + name)
            + X("wkccbd+wknccbd") * P("wkcbd_" + name)
            + X("wkempden") * P("wkempden_" + name)
            + P("ASC_" + name)
        )

    m.utility_ca = (
        +X("totcost/hhinc") * P("costbyincome")
        + X("tottime * (altid <= 4)") * P("motorized_time")
        + X("tottime * (altid >= 5)") * P("nonmotorized_time")
        + X("ovtt/dist * (altid <= 4)") * P("motorized_ovtbydist")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 17B, Segmented for 2 or more cars"

    m.choice_avail_summary()

    m.set_cap(25)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, options={"maxiter": 1000, "ftol": 1e-10})
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "Household",
            "hhinc.*",
            "vehbywrk.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    revealed_x

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -3.131353684453019,
        "ASC_SR2": -1.982918109708732,
        "ASC_SR3+": -3.7260108180838087,
        "ASC_Transit": -2.156692021346743,
        "ASC_Walk": -1.5835989189276731,
        "costbyincome": -0.0978607170004987,
        "hhinc#4": 0.0003613865368074286,
        "hhinc#5": -0.002650844116790726,
        "hhinc#6": 0.0009101590517897486,
        "motorized_ovtbydist": -0.1937603871636971,
        "motorized_time": -0.018754569843429904,
        "nonmotorized_time": -0.045210350568177106,
        "vehbywrk_Bike": -0.21719450776227095,
        "vehbywrk_SR": -0.23823848663321495,
        "vehbywrk_Transit": -0.23619743325813528,
        "vehbywrk_Walk": -0.07998140457786423,
        "wkcbd_Bike": 0.5651962301019509,
        "wkcbd_SR2": 0.16357641468531287,
        "wkcbd_SR3+": 1.3328985605250347,
        "wkcbd_Transit": 1.2613977817336346,
        "wkcbd_Walk": 0.21594406473797045,
        "wkempden_Bike": 0.0014358656965858384,
        "wkempden_SR2": 0.0010739771398805848,
        "wkempden_SR3+": 0.0013450278303126933,
        "wkempden_Transit": 0.0028932086005130325,
        "wkempden_Walk": -0.0010607308832536778,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 2e-2
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



from __future__ import annotations

import jax  # noqa: F401
import pandas as pd
from pytest import approx, fixture

import larch as lx
from larch import PX


@fixture(scope="module")
def raw_data() -> pd.DataFrame:
    return pd.read_parquet(lx.example_file("vehicle_choice.parquet"))


@fixture
def simple_model(raw_data: pd.DataFrame) -> lx.Model:
    data = lx.Dataset.construct.from_idca(raw_data)
    simple = lx.Model(data)
    simple.utility_ca = (
        PX("price")
        + PX("opcost")
        + PX("max_range")
        + PX("ev")
        + PX("hybrid")
        + PX("hiperf")
        + PX("medhiperf")
    )
    simple.choice_ca_var = "chosen"
    return simple


def test_vehicle_choice(simple_model: lx.Model):
    simple = simple_model
    simple.maximize_loglike(stderr=True, options={"ftol": 1e-9})
    assert simple.most_recent_estimation_result["loglike"] == approx(-1399.1932)
    expected_value = {
        "price": -0.4167 / 10000,
        "opcost": -0.1288 / 10,
        "max_range": 0.4770,
        "ev": -1.3924,
        "hybrid": 0.3555,
        "hiperf": 0.1099,
        "medhiperf": 0.3841,
    }
    assert simple.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=1e-3
    )
    expected_stderr = {
        "price": 0.0332 / 10000,
        "opcost": 0.0353 / 10,
        "max_range": 0.1765,
        "ev": 0.2766,
        "hybrid": 0.1218,
        "hiperf": 0.0838,
        "medhiperf": 0.0855,
    }
    assert simple.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=2e-3
    )


def test_mixed_logit(simple_model: lx.Model):
    mixed = simple_model.copy()
    mixed.mixtures = [
        lx.mixtures.Normal(k, f"s_{k}")
        for k in ["opcost", "max_range", "ev", "hybrid", "hiperf", "medhiperf"]
    ]
    mixed.n_draws = 200
    mixed.seed = 42
    assert mixed.pvals == approx(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.001, 0.001, 0.001, 0.001, 0.001, 0.001]
    )
    mixed.pvals = "null"
    assert mixed.pvals == approx(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
    )
    assert mixed.compute_engine == "jax"
    assert mixed.n_draws == 200
    assert mixed.seed == 42
    assert len(mixed.mixtures) == 6
    assert isinstance(mixed.mixtures[0], lx.mixtures.Normal)
    assert mixed.mixtures[0].to_dict() == {
        "type": "Normal",
        "mean": "opcost",
        "std": "s_opcost",
    }
    mixed.maximize_loglike(stderr=True, options={"ftol": 1e-9})
    # TEST
    assert mixed.most_recent_estimation_result["loglike"] == approx(-1385.4436)
    expected_value = {
        "ev": -3.116113749980346,
        "hiperf": 0.16526011648413752,
        "hybrid": 0.5468359000231515,
        "max_range": 0.893907686323169,
        "medhiperf": 0.7444412811439853,
        "opcost": -0.022305055541141636,
        "price": -6.53728508122224e-05,
        "s_ev": 2.8852968025136363,
        "s_hiperf": 0.7816346849525359,
        "s_hybrid": -1.2110724172285334,
        "s_max_range": -0.10080141887433067,
        "s_medhiperf": -1.5364071517526834,
        "s_opcost": 0.04830038015701635,
    }
    assert mixed.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=5e-2
    )
    expected_stderr = {
        "ev": 0.8305132389068604,
        "hiperf": 0.12845700979232788,
        "hybrid": 0.20117034018039703,
        "max_range": 0.33233582973480225,
        "medhiperf": 0.21790540218353271,
        "opcost": 0.007205564994364977,
        "price": 1.0975929399137385e-05,
        "s_ev": 0.7865121364593506,
        "s_hiperf": 0.8348018527030945,
        "s_hybrid": 0.6305373907089233,
        "s_max_range": 0.9989855885505676,
        "s_medhiperf": 0.6614137887954712,
        "s_opcost": 0.0211756881326437,
    }
    assert mixed.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=5e-2
    )

    # make sure we can copy the model
    mixed2 = mixed.copy()
    assert mixed2.compute_engine == "jax"
    assert mixed2.n_draws == 200
    assert mixed2.seed == 42
    assert len(mixed2.mixtures) == 6
    assert isinstance(mixed2.mixtures[0], lx.mixtures.Normal)
    assert mixed2.mixtures[0].to_dict() == {
        "type": "Normal",
        "mean": "opcost",
        "std": "s_opcost",
    }
    assert mixed.parameters["value"].to_series().to_dict() == approx(
        expected_value, rel=5e-2
    )
    assert mixed.parameters["std_err"].to_series().to_dict() == approx(
        expected_stderr, rel=5e-2
    )



import os

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.feather as pf
import yaml

from .aster import Asterize


def _getsize(pa_array):
    if pa_array.null_count:
        return 1_000_000
        # raise ValueError("getsize does not play nicely with nulls yet")
    try:
        return pa_array.nbytes * 8 // pa_array.type.bit_width
    except ValueError:
        # Non-fixed width type
        return 1_000_000


class Table:
    def __new__(cls, init, *args, **kwargs):
        if init is None:
            return None
        if isinstance(init, cls):
            return init
        return super().__new__(cls)

    def __init__(self, init, nthreads=None):
        if isinstance(init, Table):
            self._table = init._table
        elif isinstance(init, pd.DataFrame):
            self._table = pa.Table.from_pandas(init, nthreads=nthreads)
        elif isinstance(init, dict):
            self._table = pa.Table.from_pydict(init)
        elif isinstance(init, pa.Table):
            self._table = init
        elif isinstance(init, int):
            self._table = pa.table([])
            self._tentative_len = init
        else:
            raise TypeError(f"init is invalid type {type(init)}")
        self._revise = Asterize()

    def __repr__(self):
        r = repr(self._table)
        return r.replace("pyarrow.Table", "sharrow.Table")

    def append_column(self, name, value, overwrite=True):
        """
        Append a column to the Table.

        Unlike a regular pyarrow Table, by default this will remove any
        existing column with the same name.

        Parameters
        ----------
        name : str
        value : array-like
        overwrite : bool, default True

        Returns
        -------
        self
        """
        if len(self._table):
            if overwrite:
                try:
                    position = self._table.column_names.index(name)
                except ValueError:
                    pass
                else:
                    self._table = self._table.remove_column(position)
            self._table = self._table.append_column(name, value)
        else:
            self._table = pa.table({name: value})
        return self

    def __getattr__(self, item):
        return getattr(self._table, item)

    def __getitem__(self, item):
        return self._table[item]

    def __setitem__(self, item, value):
        if not isinstance(value, np.ndarray):
            value = np.asarray(value)
        result = np.broadcast_to(value, (self._table.num_rows))
        self._table = self._table.append_column(item, [result])

    def __len__(self):
        return len(self._table) or getattr(self, "_tentative_len", 0)

    def concat(self, others):
        tables = [self._table]
        if isinstance(others, Table):
            others = [others]
        for other in others:
            if isinstance(other, Table):
                other = other._table
            tables.append(other)
        self._table = pa.concat_tables(tables)

    @classmethod
    def from_pydict(cls, mapping):
        return cls(mapping)

    def to_feather(self, dest):
        """
        Write this table to Feather format.

        Parameters
        ----------
        dest : str
            Local destination path.
        compression : string, default None
            Can be one of {"zstd", "lz4", "uncompressed"}. The default of None uses
            LZ4 for V2 files if it is available, otherwise uncompressed.
        compression_level : int, default None
            Use a compression level particular to the chosen compressor. If None
            use the default compression level
        chunksize : int, default None
            For V2 files, the internal maximum size of Arrow RecordBatch chunks
            when writing the Arrow IPC file format. None means use the default,
            which is currently 64K
        version : int, default 2
            Feather file version. Version 2 is the current. Version 1 is the more
            limited legacy format
        """
        import pyarrow.feather as pf

        return pf.write_feather(
            self._table,
            dest,
            compression=None,
            compression_level=None,
            chunksize=None,
            version=2,
        )

    @classmethod
    def from_feather(cls, source, columns=None, memory_map=True):
        import pyarrow.feather as pf

        t = pf.read_table(source, columns=columns, memory_map=memory_map)
        return cls(t)

    def __or__(self, other):
        result = self.__class__(self._table)
        for k in other.column_names:
            result.append_column(k, other[k])
        return result

    def __dir__(self):
        return dir(self._table) + [
            "eval",
            "concat",
        ]

    def eval(
        self, expression, inplace=False, target=None, local_dict=None, nopython=True
    ):
        import numexpr as ne

        if isinstance(expression, str):
            try:
                num = float(expression.split("#")[0].strip())
            except:  # noqa: E722
                pass
            else:
                if not inplace:
                    return np.full(len(self), num)
                else:
                    if not target:
                        raise ValueError("cannot assign inplace without a target name")
                    self._table = self._table.append_column(
                        target, np.full(len(self), num)
                    )
                    return
            if target is not False:
                ex_target, ex_value = self._revise(expression)
                target = target or ex_target
            else:
                ex_value = expression
            if inplace and not target:
                target = expression
                # raise ValueError("cannot operate inplace without an assignment or target name")
            try:
                if self._table.num_rows > self._table.num_columns * 1000:
                    table = self._table
                    chunk_start = 0
                    result_chunks = []
                    while chunk_start < len(table):
                        chunk_len = min(
                            [
                                _getsize(i.slice(chunk_start).chunks[0])
                                for i in table.itercolumns()
                            ]
                        )
                        if chunk_start == 0:
                            j = ne.evaluate(
                                ex_value,
                                self._table.slice(chunk_start, chunk_len),
                                local_dict or globals(),
                            )
                        else:
                            j = ne.re_evaluate(
                                self._table.slice(chunk_start, chunk_len),
                            )
                        result_chunks.append(j)
                        chunk_start = chunk_start + chunk_len
                    expression_val = pa.chunked_array(result_chunks)
                else:
                    # wide or small tables, just go for broke
                    expression_val = ne.evaluate(
                        ex_value,
                        self._table,
                        local_dict or globals(),
                    )
            except Exception as err:
                if nopython:
                    raise type(err)(f"({expression}) {err!s}") from err
                expression_val = eval(
                    ex_value,
                    self._globals,
                    self._table,
                )
            if not inplace:
                return np.asarray(expression_val)
            try:
                if not isinstance(expression_val, pa.ChunkedArray):
                    expression_val = [expression_val]
                self._table = self._table.append_column(target, expression_val)
            except Exception as err:
                raise type(err)(f"({expression}) {err!s}") from err
        else:
            if not target:
                raise ValueError("cannot assign inplace without a target name")
            if not isinstance(expression, np.ndarray):
                expression = np.asarray(expression)
            try:
                result = np.broadcast_to(expression, (self._table.num_rows))
                if not inplace:
                    return result
                self._table = self._table.append_column(target, [result])
            except Exception as err:
                raise type(err)(f"target={target}") from err

    @classmethod
    def from_quilt(cls, path, blockname=None):
        if not os.path.isdir(path):
            raise NotADirectoryError(path)
        if blockname is not None:
            if isinstance(blockname, int):
                stopper = blockname
            else:
                qlog = os.path.join(path, "quilt.log")
                with open(qlog) as logreader:
                    existing_info = yaml.safe_load(logreader)
                for _stopper, block in enumerate(existing_info):
                    if block.get("name", None) == blockname:
                        break
                else:
                    raise KeyError(blockname)
        else:
            stopper = 1e99
        n = 0

        def rowfile(n):
            return os.path.join(path, f"block.{n:03d}.rows")

        def colfile(n):
            return os.path.join(path, f"block.{n:03d}.cols")

        builder = None
        look = True
        while look and n <= stopper:
            look = False
            if os.path.exists(rowfile(n)):
                i = pf.read_table(rowfile(n))
                if builder is None:
                    builder = cls(i)
                else:
                    builder.concat([i])
                look = True
            if os.path.exists(colfile(n)):
                i = pf.read_table(colfile(n))
                if builder is None:
                    builder = cls(i)
                else:
                    builder = builder | i
                look = True
            if look:
                n += 1
        if builder is not None:
            metadata = builder.schema.metadata
            metadata[b"quilt_number"] = f"{n}".encode()
            return builder.replace_schema_metadata(metadata)
        return None

    def to_quilt(self, path, blockname=None):
        if not os.path.exists(path):
            os.makedirs(path)
        qlog = os.path.join(path, "quilt.log")

        if not os.path.exists(qlog):
            ex_rows = 0
            ex_cols = []
            max_block = -1
        else:
            with open(qlog) as logreader:
                existing_info = yaml.safe_load(logreader)
            ex_rows = sum(block.get("rows", 0) for block in existing_info)
            ex_cols = sum((block.get("cols", []) for block in existing_info), [])
            max_block = max(block["block"] for block in existing_info)

        with open(qlog, "a") as f:
            quilt_number = max_block + 1
            f.write(f"- block: {quilt_number}\n")
            if blockname is not None:
                f.write(f"  name: {blockname}\n")
            if quilt_number == 0:
                blob_file = f"block.{quilt_number:03d}.rows"
                pf.write_feather(
                    self._table,
                    os.path.join(path, blob_file),
                )
                f.write(f"  rows: {len(self)}\n")
                f.write("  cols:\n")
                for ci in self.column_names:
                    f.write(f"    - {ci}\n")
            else:
                add_rows = self[ex_rows:]
                if len(add_rows):
                    if len(ex_cols) == 0:
                        row_block = add_rows
                    else:
                        c1 = [i for i in self.column_names if i in ex_cols]
                        row_block = add_rows.select(c1)
                    blob_file = f"block.{quilt_number:03d}.rows"
                    pf.write_feather(
                        row_block,
                        os.path.join(path, blob_file),
                    )
                    f.write(f"  rows: {len(add_rows)}\n")
                c2 = [i for i in self.column_names if i not in ex_cols]
                if len(c2):
                    col_block = self.select(c2)
                    blob_file = f"block.{quilt_number:03d}.cols"
                    pf.write_feather(
                        col_block,
                        os.path.join(path, blob_file),
                    )
                    f.write("  cols:\n")
                    for ci in c2:
                        f.write(f"    - {ci}\n")


def concat_tables(tables):
    _tables = []
    for t in tables:
        if isinstance(t, Table):
            t = t._table
        _tables.append(t)
    return Table(pa.concat_tables(_tables))



def example(extract='m', estimate=False):
    import pandas as pd

    import larch as lx

    raw_data = pd.read_csv(lx.example_file("swissmetro.csv.gz")).rename_axis(index="CASEID")
    data = lx.Dataset.construct.from_idco(raw_data, alts={1: "Train", 2: "SM", 3: "Car"})
    data

    m = lx.Model(data.dc.query_cases("PURPOSE in (1,3) and CHOICE != 0"))

    m.title = "swissmetro example 02 (weighted logit)"

    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }
    m.choice_co_code = "CHOICE"

    m.weight_co_var = "1.0*(GROUP==2)+1.2*(GROUP==3)"

    from larch import P, X

    m.utility_co[1] = P("ASC_TRAIN")
    m.utility_co[2] = 0
    m.utility_co[3] = P("ASC_CAR")
    m.utility_co[1] += X("TRAIN_TT") * P("B_TIME")
    m.utility_co[2] += X("SM_TT") * P("B_TIME")
    m.utility_co[3] += X("CAR_TT") * P("B_TIME")
    m.utility_co[1] += X("TRAIN_CO*(GA==0)") * P("B_COST")
    m.utility_co[2] += X("SM_CO*(GA==0)") * P("B_COST")
    m.utility_co[3] += X("CAR_CO") * P("B_COST")

    m.ordering = [
        (
            "ASCs",
            "ASC.*",
        ),
        (
            "LOS",
            "B_.*",
        ),
    ]
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    m.set_cap(15)
    result = m.maximize_loglike(method="SLSQP")
    result

    m.calculate_parameter_covariance()
    m.parameter_summary()
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import math

import numba as nb
import numpy as np
import pandas as pd
import xarray as xr

try:
    import sparse
except ImportError:
    sparse = None

try:
    import scipy.sparse
except ImportError:
    scipy = None


@nb.njit
def _get_idx(indices, indptr, data, i, j):
    pool_lb = indptr[i]
    pool_ub = indptr[i + 1] - 1
    if pool_ub < pool_lb:
        # This indicates there are no values at all for row i
        return np.nan
    idx_lo = indices[pool_lb]
    idx_hi = indices[pool_ub]
    # check top and bottom
    if j == idx_lo:
        # The lower bound on possible j values is the j value, return it
        return data[pool_lb]
    elif j == idx_hi:
        # The upper bound on possible j values is the j value, return it
        return data[pool_ub]
    # check if out of original range
    elif j < idx_lo or j > idx_hi:
        # the j value is outside the possible range, there is no value to return
        return np.nan

    # The j value is somewhere inside the bounds, so conduct an efficient search to
    # see if we can find it.
    span = pool_ub - pool_lb - 1
    # assume the j values are uniformly distributed between bounds, guess at the
    # approximate location of the target j
    peek = (j - idx_lo) / (idx_hi - idx_lo)
    while span > 3:
        candidate = int(peek * span) + pool_lb
        if candidate <= pool_lb:
            candidate = pool_lb + 1
        if candidate >= pool_ub:
            candidate = pool_ub - 1
        # printd(f"{peek=}  {span=}  {pool_lb=}  {int(peek * span)=}  {candidate=}  ???")
        if j == indices[candidate]:
            # printd(f"{pool_lb=}  {pool_ub=}  {indices[pool_lb]=}  {indices[pool_ub]=}  {span=}  {i=}  {j=} *")
            return data[candidate]
        elif j < indices[candidate]:
            pool_ub = candidate
            idx_hi = indices[pool_ub]
            span = pool_ub - pool_lb - 1
            # printd(f"{pool_lb=}  {pool_ub=}  {indices[pool_lb]=}  {indices[pool_ub]=}  {span=}  {i=}  {j=} <")
        elif j > indices[candidate]:
            pool_lb = candidate
            idx_lo = indices[pool_lb]
            span = pool_ub - pool_lb - 1
            # printd(f"{pool_lb=}  {pool_ub=}  {indices[pool_lb]=}  {indices[pool_ub]=}  {span=}  {i=}  {j=} >")
        peek = (j - idx_lo) / (idx_hi - idx_lo)
    while pool_lb < pool_ub:
        pool_lb += 1
        if j == indices[pool_lb]:
            # printd(f"{pool_lb=}  {pool_ub=}  {indices[pool_lb]=}  {indices[pool_ub]=}  {span=}  {i=}  {j=} *")
            return data[pool_lb]
    # printd("no match")
    return np.nan


class SparseArray2D:
    def __init__(self, i, j, data, shape=None):
        if scipy is None:
            raise ImportError("scipy is not installed")
        if isinstance(data, scipy.sparse.csr_matrix):
            self._sparse_data = data
        else:
            self._sparse_data = scipy.sparse.coo_matrix(
                (data, (i, j)), shape=shape
            ).tocsr()
        self._sparse_data.sort_indices()

    def __getitem__(self, item):
        if isinstance(item, tuple) and len(item) == 2:
            if isinstance(item[0], int) and isinstance(item[1], int):
                return _get_idx(
                    self._sparse_data.indices,
                    self._sparse_data.indptr,
                    self._sparse_data.data,
                    item[0],
                    item[1],
                )
        return type(self)(None, None, self._sparse_data[item])


@xr.register_dataset_accessor("redirection")
class RedirectionAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def set(self, m2t, map_to, map_also=None, name=None):
        """
        Set the redirection of a dimension.

        Parameters
        ----------
        m2t : pandas.Series
            Mapping maz's to tazs
        """
        if name is None:
            name = f"redirect_{map_to}"

        mapper = {i: j for (j, i) in enumerate(self._obj[map_to].to_numpy())}
        if isinstance(m2t, pd.DataFrame) and m2t.shape[1] == 1:
            m2t = m2t.iloc[:, 0]
        if isinstance(m2t, pd.Series):
            m2t = xr.DataArray(m2t, dims=name)
        offsets = xr.apply_ufunc(np.vectorize(mapper.get), m2t)
        self._obj[f"_digitized_{map_to}_of_{name}"] = offsets
        self._obj.attrs[f"dim_redirection_{name}"] = map_to

        if map_also is not None:
            for i, j in map_also.items():
                self._obj[f"_digitized_{i}_of_{j}"] = offsets.rename({name: j})
                self._obj.attrs[f"dim_redirection_{j}"] = i

    def sparse_blender(
        self,
        name,
        i,
        j,
        data,
        shape=None,
        max_blend_distance=None,
        blend_distance_name=None,
        index=None,
        i_dim="omaz",
        j_dim="dmaz",
    ):
        i = np.asarray(i)
        j = np.asarray(j)
        if index is not None and shape is None:
            shape = (len(index), len(index))
        if index is not None:
            mapper = {i: j for (j, i) in enumerate(index)}

            @np.vectorize
            def apply_mapper(x):
                return mapper.get(x, -1)

            i_ = apply_mapper(i)
            j_ = apply_mapper(j)
        else:
            i_ = i
            j_ = j

        if sparse is None:
            raise ImportError("sparse is not installed")

        sparse_data = sparse.GCXS(
            sparse.COO(np.stack((i_, j_)), data, shape=shape), compressed_axes=(0,)
        )
        self._obj[f"_s_{name}"] = xr.DataArray(
            sparse_data,
            dims=(i_dim, j_dim),
        )
        if not max_blend_distance:
            max_blend_distance = np.inf
        self._obj.attrs[f"blend_{name}_max"] = max_blend_distance
        self._obj.attrs[f"blend_{name}_dist"] = blend_distance_name

    def is_blended(self, name):
        return f"_s_{name}" in self._obj
        # return (
        #     (f"_{name}_indices" in self._obj)
        #     and (f"_{name}_indptr" in self._obj)
        #     and (f"_{name}_data" in self._obj)
        # )

    def get_blended(self, name, backstop_values, i, j):
        return get_blended_2_arr(
            backstop_values,
            np.asarray(self._obj[f"_s_{name}"].data.indices),
            np.asarray(self._obj[f"_s_{name}"].data.indptr),
            np.asarray(self._obj[f"_s_{name}"].data.data),
            i,
            j,
            blend_limit=self._obj.attrs[f"blend_{name}_max"],
        )

    def target(self, name):
        return self._obj.attrs.get(f"dim_redirection_{name}", None)

    @property
    def blenders(self):
        b = {}
        for k, v in self._obj.attrs.items():
            if k.startswith("blend_") and k.endswith("_max"):
                name = k[6:-4]
                d = b.get(name, {})
                d["max_blend_distance"] = v
                b[name] = d
            if k.startswith("blend_") and k.endswith("_dist"):
                name = k[6:-5]
                d = b.get(name, {})
                d["blend_distance_name"] = v
                b[name] = d
        return b


# fastmath must be false to ensure NaNs are detected here.
# wrapping this as such allows fastmath to be turned on in outer functions
# but not lose the ability to check for NaNs.  Older versions of this function
# checked whether the float cast to an integer was -9223372036854775808, but
# that turns out to be not compatible with all hardware (i.e. Apple Silicon).
def isnan_fast_safe(x):
    if isinstance(x, float):
        return math.isnan(x)
    elif isinstance(x, str):
        return x == "\u0015"
    else:
        return False


@nb.extending.overload(isnan_fast_safe, jit_options={"fastmath": False})
def ol_isnan_fast_safe(x):
    if isinstance(x, nb.types.Float):

        def func(x):
            return math.isnan(x)

        return func
    elif isinstance(x, (nb.types.UnicodeType, nb.types.UnicodeCharSeq)):
        return lambda x: x == "\u0015"
    else:
        return lambda x: False


@nb.njit
def get_blended_2(backstop_value, indices, indptr, data, i, j, blend_limit=np.inf):
    dtype = type(backstop_value)
    micro_v = dtype(_get_idx(indices, indptr, data, i, j))
    if isnan_fast_safe(micro_v) or micro_v > blend_limit:
        return backstop_value
    if blend_limit == np.inf:
        return micro_v
    macro_ratio = dtype(micro_v / blend_limit)
    return macro_ratio * backstop_value + (1 - macro_ratio) * micro_v


@nb.njit
def get_blended_2_arr(
    backstop_values_, indices, indptr, data, i_, j_, blend_limit=np.inf
):
    out = np.zeros_like(backstop_values_)
    for z in range(backstop_values_.size):
        out[z] = get_blended_2(
            backstop_values_[z],
            indices,
            indptr,
            data,
            i_[z],
            j_[z],
            blend_limit=blend_limit,
        )
    return out



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .relationships import DataTree


class DataTreeBranch:
    """Access variables in a subspace of a tree.

    This class enables the use of the python and numexpr engines to access
    dotted data in the tree.
    """

    def __init__(self, tree: DataTree, branch: str):
        self.tree = tree
        self.branch = branch

    def __getitem__(self, key):
        return self.tree[self.branch + "." + key]

    def __setitem__(self, key, value):
        self.tree[self.branch + "." + key] = value

    def __getattr__(self, item):
        dataset = self.tree.subspaces[self.branch]
        if item in dataset:
            return self.tree[self.branch + "." + item]
        else:
            raise AttributeError(f"{item} not found in {self.branch}")


class CachedTree:
    """A wrapper that caches the results of getitem calls."""

    def __init__(self, tree):
        self._tree = tree
        self._cache = {}

    def __getitem__(self, item):
        x = self._cache.get(item, None)
        if x is None:
            x = self._cache[item] = self._tree[item]
        return x
